import tkinter as tk
from tkinter import filedialog
from tkinter import messagebox
from functions import *
import sys

doc = f"""Bienvenue sur le Trieur d'amendements DLF!

Pour utiliser cet outil, veuillez suivre les étapes suivantes:

1. Indiquez l'emplacement de la moulinette, du PLF à trier et de l'historique des PLFs en cliquant sur les boutons 'Parcourir...'.

2. Définissez le score de ressemblance souhaité entre les amendements du PLF à trier et ceux de l'historique des PLFs en déplaçant le curseur.

3. Cliquez sur le bouton 'Trier les amendements' pour lancer le tri.

4. Le résultat sera affiché dans une boîte de dialogue.

Si vous avez besoin d'aide, n'hésitez pas à cliquer sur le bouton '?' pour afficher cette documentation.

Inputs:
- Moulinette location: emplacement de la moulinette déjà lancée manuellement
- PLF à trier location: emplacement des liasses du PLF à trier
- Historique des PLFs location: emplacement de l'historique des PLFs précédents au format xlsx
- Score de ressemblance: score de ressemblance souhaité entre les amendements

Bonne utilisation!
"""

class AideWindow:
    def __init__(self, master):
        self.master = master
        master.title("Aide")
        self.label = tk.Label(master, text=doc)
        self.label.pack()

class TrieurAmendementsDLF:
    def __init__(self, master):
        self.master = master
        master.title("Trieur d'amendements DLF")

        # Input pour la moulinette
        self.label_moulinette = tk.Label(master, text="Moulinette déjà lancée:")
        self.label_moulinette.grid(row=0, column=0)
        self.entry_moulinette = tk.Entry(master)
        self.entry_moulinette.grid(row=0, column=1)
        self.button_moulinette = tk.Button(master, text="Parcourir...", command=self.get_moulinette_location)
        self.button_moulinette.grid(row=0, column=2)

        # Input pour le PLF à trier
        self.label_plf_a_trier = tk.Label(master, text="Liasses PLF à trier:")
        self.label_plf_a_trier.grid(row=1, column=0)
        self.entry_plf_a_trier = tk.Entry(master)
        self.entry_plf_a_trier.grid(row=1, column=1)
        self.button_plf_a_trier = tk.Button(master, text="Parcourir...", command=self.get_plf_a_trier_location)
        self.button_plf_a_trier.grid(row=1, column=2)

        # Input pour l'historique des PLFs
        self.label_historique_plfs = tk.Label(master, text="Historique des PLFs:")
        self.label_historique_plfs.grid(row=2, column=0)
        self.entry_historique_plfs = tk.Entry(master)
        self.entry_historique_plfs.grid(row=2, column=1)
        self.button_historique_plfs = tk.Button(master, text="Parcourir...", command=self.get_historique_plfs_location)
        self.button_historique_plfs.grid(row=2, column=2)

        # Curseur pour le score de ressemblance
        self.score_ressemblance = tk.IntVar(value=80)
        self.label_score_ressemblance = tk.Label(master, text="Score de ressemblance PLF actuel:")
        self.label_score_ressemblance.grid(row=3, column=0)
        self.scale_score_ressemblance = tk.Scale(master, from_=1, to=100, orient=tk.HORIZONTAL,length=200,variable=self.score_ressemblance)
        self.scale_score_ressemblance.grid(row=3, column=1)
        
         # Curseur pour le score de ressemblance
        self.score_ressemblance2 = tk.IntVar(value=70)
        self.label_score_ressemblance2 = tk.Label(master, text="Score de ressemblance PLF passés:")
        self.label_score_ressemblance2.grid(row=4, column=0)
        self.scale_score_ressemblance2 = tk.Scale(master, from_=1, to=100, orient=tk.HORIZONTAL,length=200,variable=self.score_ressemblance2)
        self.scale_score_ressemblance2.grid(row=4, column=1)

        # Bouton pour trier les amendements
        self.button_trier = tk.Button(master, text="Trier les amendements", command=self.trier_amendements)
        self.button_trier.grid(row=5, column=1)

        # Case besoin d'aide
        self.label_aide = tk.Label(master, text="Besoin d'aide?")
        self.label_aide.grid(row=6, column=0)
        self.button_aide = tk.Button(master, text="?", command=self.afficher_aide)
        self.button_aide.grid(row=6, column=1)

    def get_moulinette_location(self):
        location = filedialog.askopenfilename()
        self.entry_moulinette.insert(0, location)

    def get_plf_a_trier_location(self):
        location = filedialog.askdirectory()
        self.entry_plf_a_trier.insert(0, location)

    def get_historique_plfs_location(self):
        location = filedialog.askdirectory()
        self.entry_historique_plfs.insert(0, location)

    def trier_amendements(self):
        #sys.stdout = open('./logs.txt', 'w')
        
        try:
            os.remove("../logs.txt")
        except OSError:
            pass

    
        moulinette_location = self.entry_moulinette.get()
        plf_a_trier_location = self.entry_plf_a_trier.get()
        historique_plfs_location = self.entry_historique_plfs.get()
        score_ressemblance = self.scale_score_ressemblance.get()
        score_ressemblance2 = self.scale_score_ressemblance2.get()
        
        functions_mapping = {
            "AN": {
                "update_function": get_retires_and_inexistants,
                "extraction_function": get_corps_amendements,
                "extraction_place_function": get_place,
                "comparison_function": compare_same_PLF,
                "comparison_function_old": compare_old_PLF
            },
            "SENAT": {
                "update_function": get_retires_and_inexistants_senat,
                "extraction_function": get_corps_amendements_senat,
                "extraction_place_function": get_place_senat,
                "comparison_function": compare_same_PLF_senat,
                "comparison_function_old": compare_old_PLF_senat
            }
        }
        
        
        print("###############\n"*5)
        print("on charge les amendements")
        print("###############\n"*5)
        d,doc_type = identifier_document_type(plf_a_trier_location)
        doc_functions = functions_mapping.get(doc_type, None)
        if doc_functions:
            update_function = doc_functions.get("update_function")
            extraction_function = doc_functions.get("extraction_function")
            extraction_place_function = doc_functions.get("extraction_place_function")
            comparison_function = doc_functions.get("comparison_function")
            comparison_function_old = doc_functions.get("comparison_function_old")
        
        moulinette2 = pd.read_excel(moulinette_location,sheet_name=1)
        moulinette2[moulinette2.columns[0]] = pd.to_numeric(moulinette2[moulinette2.columns[0]].replace('[a-zA-Z]', '', regex=True))
        moulinette2 = moulinette2[moulinette2["N ° amdt"] != 0]
        nattributed = moulinette2[~moulinette2["Attribution finale"].notnull()]
        attributed = moulinette2[moulinette2["Attribution finale"].notnull()]
        get_attrib_score(attributed,nattributed)
        
        print("###############\n"*5)
        print("on enlève les amendements qui ont été retirés avant publication et ceux qui sont inexistants")
        print("###############\n"*5)
        
        # Application de la fonction choisie pour mettre à jour 'moulinette2', si un type de document valide est fourni
        if update_function:
            moulinette2 = update_function(nattributed, moulinette2, d)
        nattributed = moulinette2[~moulinette2["Attribution finale"].notnull()]
        attributed = moulinette2[moulinette2["Attribution finale"].notnull()]
        get_attrib_score(attributed,nattributed)
        
        print("###############\n"*5)
        print("on récupère le corps des amendements")
        print("###############\n"*5)
        # Extraction des corps des amendements en utilisant la fonction appropriée
        if extraction_function:
            corps_dict, corps_dict2, corps_dict3 = extraction_function(d)
        mapping_PLF = mapping_PLF_function(corps_dict2)
        write_PLF(corps_dict3)
          
        print("###############\n"*5)
        print("Attribution de l’amendement par l’IA (en fonction des mots clés relevés dans le corps")
        print("###############\n"*5)
        moulinettePLF, moulinetteCGI, moulinetteCIBS, moulinetteCDD, moulinetteCodeExt, moulinetteMots = (pd.read_excel(moulinette_location, sheet_name=sheet_num) for sheet_num in [2, 3, 4, 5, 6, 7])
        CGI,PLF,CIBS,CDD,CodeExt,Mots = get_dicts(moulinetteCGI,moulinettePLF,moulinetteCIBS,moulinetteCDD,moulinetteCodeExt,moulinetteMots)

        # Extraction des places en utilisant la fonction appropriée
        if extraction_place_function:
            place_dict = extraction_place_function(d, nattributed)

        moulinette2 = attribution_via_texte(moulinette2,attributed,nattributed,corps_dict,corps_dict2,score_ressemblance,CGI,CDD,CIBS,PLF,CodeExt,Mots,place_dict)
        nattributed = moulinette2[~moulinette2["Attribution finale"].notnull()]
        attributed = moulinette2[moulinette2["Attribution finale"].notnull()]
        get_attrib_score(attributed,nattributed)
        
        print("###############\n"*5)
        print("identifier les numéros d’amendements identiques au sein d’une même lecture")
        print("###############\n"*5)
        if comparison_function:
            moulinette2 = comparison_function(corps_dict2, mapping_PLF, nattributed, attributed, moulinette2, score_ressemblance)

        nattributed = moulinette2[~moulinette2["Attribution finale"].notnull()]
        attributed = moulinette2[moulinette2["Attribution finale"].notnull()]
        get_attrib_score(attributed,nattributed)
        
        print("###############\n"*5)
        print("Identification des identiques dans les PLF précédents")
        print("###############\n"*5) 
        liste_doc = glob.glob(historique_plfs_location+"\\*.xlsx")
        df,mapping_historique = get_old_PLFs(liste_doc)
        
        if comparison_function:
            moulinette2 = comparison_function_old(df, corps_dict2, mapping_PLF, nattributed, attributed, moulinette2, score_ressemblance2, mapping_historique)
       
        nattributed = moulinette2[~moulinette2["Attribution finale"].notnull()]
        attributed = moulinette2[moulinette2["Attribution finale"].notnull()]
        get_attrib_score(attributed,nattributed)
        
        print("attribution supplémentaire mots clés")
        try:
            moulinetteMots_expo = pd.read_excel(moulinette_location,sheet_name=8)
            Mots_expo = get_dicts_expo_sommaire(moulinetteMots_expo)
            moulinette2 = attribution_via_mots_cles_expo_sommaire(moulinette2,attributed,nattributed,corps_dict2,corps_dict3,score_ressemblance,Mots,place_dict,Mots_expo,CodeExt)
            nattributed = moulinette2[~moulinette2["Attribution finale"].notnull()]
            attributed = moulinette2[moulinette2["Attribution finale"].notnull()]
            get_attrib_score(attributed,nattributed)
        except Exception as e:
            print(e)
            print("pas de feuille exposé sommaire")
        
        print("###############\n"*5)
        print("Codes qu'il manque au niveau de la moulinette")
        print("###############\n"*5)
        missing_code(moulinette2)
        moulinette2.to_csv("../attribution_finale.csv")
        ecrire_resultat(moulinette_location)
        tk.messagebox.showinfo("Information","Terminé, ouvrir le fichier logs.txt pour plus d'informations, bon courage à tous et toutes! Joseph")
    
    def afficher_aide(self):
        aide_window = tk.Toplevel(self.master)
        app = AideWindow(aide_window)

if __name__ == "__main__":
    root = tk.Tk()  # Création de l'instance de Tk
    app = TrieurAmendementsDLF(root)  # Création de l'instance de la classe TrieurAmendementsDLF
    root.mainloop()  # Lancement de la boucle principale d'événements de Tk